@@ -1,62 +1,72 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2007 Christopher Lenz
+# Copyright (C) 2007-2009 Christopher Lenz
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
 
+"""Utility for dumping a snapshot of a CouchDB database to a multipart MIME
+file.
+"""
+
 from base64 import b64decode
 from email.MIMEBase import MIMEBase
 from email.MIMEMultipart import MIMEMultipart
 from optparse import OptionParser
-try:
-    import simplejson as json
-except ImportError:
-    import json # Python 2.6
 import sys
 
 from couchdb import __version__ as VERSION
+from couchdb import json
 from couchdb.client import Database
+from couchdb.multipart import write_multipart
 
 
-def dump_db(dburl, username=None, password=None, boundary=None):
-    envelope = MIMEMultipart('mixed', boundary)
+def dump_db(dburl, username=None, password=None, boundary=None,
+            output=sys.stdout):
     db = Database(dburl)
     if username is not None and password is not None:
         db.resource.http.add_credentials(username, password)
+
+    envelope = write_multipart(output)
+
     for docid in db:
         doc = db.get(docid, attachments=True)
         print>>sys.stderr, 'Dumping document %r' % doc.id
         attachments = doc.pop('_attachments', {})
-
-        part = MIMEBase('application', 'json')
-        part.set_payload(json.dumps(doc, sort_keys=True, indent=2))
+        jsondoc = json.encode(doc)
 
         if attachments:
-            inner = MIMEMultipart('mixed')
-            inner.attach(part)
+            parts = envelope.open({
+                'Content-ID': doc.id,
+                'ETag': '"%s"' % doc.rev
+            })
+            parts.add('application/json', jsondoc)
+
             for name, info in attachments.items():
                 content_type = info.get('content_type')
                 if content_type is None: # CouchDB < 0.8
                     content_type = info.get('content-type')
-                maintype, subtype = content_type.split('/', 1)
-                subpart = MIMEBase(maintype, subtype)
-                subpart['Content-ID'] = name
-                subpart.set_payload(b64decode(info['data']))
-                inner.attach(subpart)
-            part = inner
+                parts.add(content_type, b64decode(info['data']), {
+                    'Content-ID': name
+                })
+            parts.close()
 
-        part['Content-ID'] = doc.id
-        part['ETag'] = doc.rev
+        else:
+            envelope.add('application/json', jsondoc, {
+                'Content-ID': doc.id,
+                'ETag': '"%s"' % doc.rev
+            }, )
 
-        envelope.attach(part)
-    return envelope.as_string()
+    envelope.close()
 
 
 def main():
     parser = OptionParser(usage='%prog [options] dburl', version=VERSION)
+    parser.add_option('--json-module', action='store', dest='json_module',
+                      help='the JSON module to use ("simplejson", "cjson", '
+                            'or "json" are supported)')
     parser.add_option('-u', '--username', action='store', dest='username',
                       help='the username to use for authentication')
     parser.add_option('-p', '--password', action='store', dest='password',
@@ -67,8 +77,10 @@ def main():
     if len(args) != 1:
         return parser.error('incorrect number of arguments')
 
-    print dump_db(args[0], username=options.username,
-                  password=options.password)
+    if options.json_module:
+        json.use(options.json_module)
+
+    dump_db(args[0], username=options.username, password=options.password)
 
 
 if __name__ == '__main__':
